#include <time.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>

#define CMP		>
#define SIZE		27
#define POOL_SIZE	5
#define LENGTH		(SIZE * SIZE)
#define TIMEOUT 	(LENGTH * LENGTH * POOL_SIZE)

#define ARRAY_SIZE(arr) (sizeof(arr) / sizeof((arr)[0]))

typedef struct {
    long score;
    int data[LENGTH];
} Layout;

static Layout legend;
static Layout champion;
static Layout pool[POOL_SIZE];

static int better(Layout *a, Layout *b) {
    return a->score CMP b->score;
}

static int diff(Layout *a, Layout *b) {
    int i;
    for (i = 0; i < LENGTH; i++) {
	if (a->data[i] != b->data[i]) return 1;
    }
    return 0;
}

static int equal(Layout *a, Layout *b) {
    return (a->score != b->score) ? 0 : !diff(a, b);
}

static void print(Layout *layout) {
    int i;
    printf("SCORE: %ld\n", layout->score);
    for (i = 0; i < LENGTH; i++) {
	if (i != 0) printf(", ");
	if (i % SIZE == 0) printf("(");
	printf("%d", layout->data[i]);
	if (i % SIZE == SIZE - 1) printf(")");
    }
    printf("\n");
    fflush(NULL);
}

static int gcd(int a, int b) {
    while (a != b) {
	if (a > b) a = a - b; else b = b - a;
    }
    return a;
}

static int square(int value) {
    return value * value;
}

static int distance(int i, int j) {
    return square(i % SIZE - j % SIZE) + square(i / SIZE - j / SIZE);
}

static int gcd_table[LENGTH + 1][LENGTH + 1];
static int distance_table[LENGTH][LENGTH];
static int center_distance[LENGTH];
static int goodness_order[LENGTH];

static float distance_to_center(int i) {
    float a = (i % SIZE) - ((SIZE - 1) / 2.0);
    float b = (i / SIZE) - ((SIZE - 1) / 2.0);
    return a * a + b * b;
}

static int compare_distance(const int *a, const int *b) {
    return distance_to_center(*b) CMP distance_to_center(*a);
}

static int goodness(int num) {
    int i, sum = 0;
    for (i = 1; i <= LENGTH; i++) {
	if (i != num && gcd(i, num) > 1) sum++;
    }
    return sum;
}

static int compare_goodness(const int *a, const int *b) {
    return goodness(*b) > goodness(*a);
}

static void precalculate_tables(void) {
    int i, j;
    for (i = 0; i < LENGTH; i++) {
	center_distance[i] = i;
	goodness_order[i] = i + 1;
	for (j = 0; j < LENGTH; j++) {
	    gcd_table[i + 1][j + 1] = gcd(i + 1, j + 1);
	    distance_table[i][j] = distance(i, j);
	}
    }
    qsort(center_distance, LENGTH, sizeof(int), (void *) &compare_distance);
    qsort(goodness_order, LENGTH, sizeof(int), (void *) &compare_goodness);
}

static int cell_value(Layout *layout, int i, int j) {
    return gcd_table[layout->data[i]][layout->data[j]] * distance_table[i][j];
}

static void delacorte(Layout *layout) {
    int i, j;
    layout->score = 0;
    for (i = 0; i < LENGTH; i++) {
	for (j = i + 1; j < LENGTH; j++) {
	    layout->score += cell_value(layout, i, j);
	}
    }
}

static int delacorte_cell(Layout *layout, int i) {
    int j, score = 0;
    for (j = 0; j < LENGTH; j++) {
	if (i != j) score += cell_value(layout, i, j);
    }
    return score;
}

static int delacorte_pair(Layout *layout, int i, int j) {
    return delacorte_cell(layout, i) + delacorte_cell(layout, j);
}

static void swap(Layout *layout, int i, int j) {
    layout->score -= delacorte_pair(layout, i, j);
    int temporary = layout->data[i];
    layout->data[i] = layout->data[j];
    layout->data[j] = temporary;
    layout->score += delacorte_pair(layout, i, j);
}

static void incremental(Layout *layout) {
    int i;
    for (i = 0; i < LENGTH; i++) {
	layout->data[center_distance[i]] = goodness_order[i];
    }
    delacorte(layout);
}

static int is_digit(char c) {
    return c >= '0' && c <= '9';
}

static void parse(char *buf, Layout *dst) {
    int i = 0, j = 0;
    while (buf[i] && j < LENGTH) {
	if (is_digit(buf[i])) {
	    dst->data[j++] = atoi(buf + i);
	    while (buf[i] && buf[i] != ',') i++;
	}
	i++;
    }
    delacorte(dst);
}

static void populate_pool(void) {
    int i;
    for (i = 0; i < POOL_SIZE; i++) {
	pool[i] = champion;
    }
    i = 0;
    char buf[8192];
    sprintf(buf, "%02d-%s.txt", SIZE, 1 CMP 0 ? "max" : "min");
    FILE *file = fopen(buf, "r");
    if (file) {
	while (fgets(buf, sizeof(buf), file)) {
	    fgets(buf, sizeof(buf), file);
	    parse(buf, pool + i);
	    legend = pool[i];
	    i = (i + 1) % POOL_SIZE;
	}
    }
}

static Layout *loser = pool;
static Layout *successful = NULL;
static void replace(Layout *layout) {
    int i;
    successful = NULL;
    if (better(layout, loser)) {
	for (i = 0; i < POOL_SIZE; i++) {
	    if (equal(layout, pool + i)) return;
	}
	*loser = *layout;
	successful = loser;
	for (i = 0; i < POOL_SIZE; i++) {
	    if (better(loser, pool + i)) {
		loser = pool + i;
	    }
	}
    }
}

static Layout *random_parent(void) {
    return successful ? successful : pool + (rand() % POOL_SIZE);
}

typedef void (*Mutation)(Layout *layout);

static int skew(void) {
    return rand() % LENGTH;
}

static void swapper(Layout *layout) {
    swap(layout, center_distance[skew()], center_distance[skew()]);
}

static Mutation mutations[] = {
    swapper,
};

static void mutate(Layout *layout) {
    *layout = *random_parent();
    mutations[rand() % ARRAY_SIZE(mutations)](layout);
}

static long ticks;

static void reset_time(void) {
    ticks = TIMEOUT;
}

static int is_timeout(void) {
    return ticks-- < 0;
}

static void randomize(Layout *layout) {
    *layout = legend;
    int j = rand() % (LENGTH - 1);
    swap(layout, center_distance[j], center_distance[j + 1]);
}

static void reset_pool(void) {
    int i;
    randomize(&champion);
    for (i = 0; i < POOL_SIZE; i++) {
	randomize(pool + i);
    }
}

static void print_champion(void) {
    if (better(&champion, &legend)) {
	legend = champion;
	print(&legend);
    }
}

static void fine_tune(void) {
    reset_time();
    for (;;) {
	Layout layout;
	mutate(&layout);
	replace(&layout);
	if (better(&layout, &champion)) {
	    champion = layout;
	    print_champion();
	    reset_time();
	}
	else if (is_timeout()) {
	    reset_pool();
	    reset_time();
	}
    }
}

int main(int argc, char **argv) {
    precalculate_tables();
    incremental(&champion);
    legend = champion;
    populate_pool();
    fine_tune();
    return 0;
}
